/************************************************************************************

   STV-TOOLS: Shannon Total Variation based image processing Copyright (C) 2016
   Remy Abergel (remy.abergel AT gmail.com), Lionel Moisan (Lionel.Moisan AT
   parisdescartes.fr).

   This program is free software: you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free Software
   Foundation, either version 3 of the License, or (at your option) any later
   version.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along with
   this program.  If not, see <http://www.gnu.org/licenses/>.

                           IMPORTANT NOTE

   Part of this code is inspired from the source file CImg.h of the CImg library
   (David Tschumperlé - GREYC UMR CNRS 6072, Image team).

   The CImg software is governed either by the CeCILL or CeCILL-C license under
   French law and abiding by the rules of distribution of free software. You can
   use, modify and or redistribute the software under the terms of the CeCILL or
   CeCILL-C licenses as circulated by CEA, CNRS and INRIA at the following URL:
   "http://www.cecill.info".

   As a counterpart to the access to the source code and rights to copy, modify
   and redistribute granted by the license, users are provided only with a
   limited warranty and the software's author, the holder of the economic
   rights, and the successive licensors have only limited liability.

   In this respect, the user's attention is drawn to the risks associated with
   loading, using, modifying and/or developing or reproducing the software by
   the user in light of its specific status of free software, that may mean that
   it is complicated to manipulate, and that also therefore means that it is
   reserved for developers and experienced professionals having in-depth
   computer knowledge. Users are therefore encouraged to load and test the
   software's suitability as regards their requirements in conditions enabling
   the security of their systems and/or data to be ensured and, more generally,
   to use and operate it in the same conditions as regards security.

   The fact that you are presently reading this means that you have had
   knowledge of the CeCILL and CeCILL-C licenses and that you accept its terms.

*************************************************************************************/

/*

  Versions :

  + v 1.0 : (09/2016) initial version (Rémy Abergel)

*/

#include <stdlib.h>
#include <stdio.h>
#include <tiffio.h>

/* prototypes */
static int _save_tiff_uint8(TIFF*,double*,int,int);
static int _save_tiff_int8(TIFF*,double*,int,int);
static int _save_tiff_float(TIFF*,double*,int,int);

int save_tiff(filename,pixmap,nx,ny,bitspersample,is_tifffloat,is_unsigned,verbose)
     char *filename,is_tifffloat,is_unsigned;
     double *pixmap;
     int nx,ny;
     uint16 bitspersample;
{
  int err,adr;
  TIFF *tiffp = TIFFOpen(filename,"w");

  if(verbose) {
    printf("save '%s' image (%s): in progress",filename,is_tifffloat ? "float precision" : "uint8 precision");
    fflush(stdout);
  }

  if (tiffp) {
    if(is_tifffloat) err = _save_tiff_float(tiffp,pixmap,nx,ny);
    else {
      switch(bitspersample){
      case 8 :
	err = (is_unsigned) ? _save_tiff_uint8(tiffp,pixmap,nx,ny) : _save_tiff_int8(tiffp,pixmap,nx,ny);
	break;
      case 16 :
	//err = (is_unsigned) ? _save_tiff_uint16(tiffp,pixmap,nx,ny) : _save_tiff_int16(tiffp,pixmap,nx,ny);
	break;
      case 32 :
	//err = (is_unsigned) ? _save_tiff_uint32(tiffp,pixmap,nx,ny) : _save_tiff_int32(tiffp,pixmap,nx,ny);
	break;
      }
    }
  } else err = EXIT_FAILURE;

  if(verbose) {
    printf("\33[2K\rsave '%s' image (%s): %s\n",filename,is_tifffloat ? "float precision" : "uint8 precision",EXIT_SUCCESS == err ? "done" : "failure");
    fflush(stdout);
  }

  return err;
}


/*
 * Functions _save_tiff_xxxx: TIFF single directory writer.
 *
 * This function writes a single directory, as a:
 * 	- Tiled, Contiguous Planar Config,
 * 	- (u)int(8,16,32) or float datatype (according to xxxx),
 * 	- monochannel or multichannel image.
 *
 * These functions returns EXIT_SUCCESS or EXIT_FAILURE integer.
 *
 * */
static int _save_tiff_uint8(tiffp,pixmap,nx,ny)
     TIFF *tiffp;
     double *pixmap;
     int nx,ny;
{
  const unsigned int directory = 0;
  unsigned int row,rr,cc,vv;
  uint32 nrow;
  tstrip_t strip;
  tsize_t i;
  uint16 photometric = PHOTOMETRIC_MINISBLACK;
  uint32 rowsperstrip = (uint32)-1;
  uint16 samplesperpixel = 1;
  uint16 compression_type = 1;
  uint16 bitspersample = 8;
  double val;

  TIFFSetDirectory(tiffp,directory);
  TIFFSetField(tiffp,TIFFTAG_IMAGEWIDTH,nx);
  TIFFSetField(tiffp,TIFFTAG_IMAGELENGTH,ny);
  TIFFSetField(tiffp,TIFFTAG_ORIENTATION,ORIENTATION_TOPLEFT);
  TIFFSetField(tiffp,TIFFTAG_SAMPLESPERPIXEL,samplesperpixel);
  TIFFSetField(tiffp,TIFFTAG_SAMPLEFORMAT,1);
  TIFFSetField(tiffp,TIFFTAG_BITSPERSAMPLE,bitspersample);
  TIFFSetField(tiffp,TIFFTAG_PLANARCONFIG,PLANARCONFIG_CONTIG);
  TIFFSetField(tiffp,TIFFTAG_PHOTOMETRIC,photometric);
  TIFFSetField(tiffp,TIFFTAG_COMPRESSION,compression_type);
  rowsperstrip = TIFFDefaultStripSize(tiffp,rowsperstrip);
  TIFFSetField(tiffp,TIFFTAG_ROWSPERSTRIP,rowsperstrip);
  TIFFSetField(tiffp,TIFFTAG_FILLORDER,FILLORDER_MSB2LSB);

  uint8 *buf = (uint8*) _TIFFmalloc(TIFFStripSize(tiffp));
  if (buf) {
    for (row = 0; row<ny; row+=rowsperstrip) {
      nrow = (row + rowsperstrip>ny?ny-row:rowsperstrip);
      strip = TIFFComputeStrip(tiffp,row,0);
      i = 0;
      for (rr = 0; rr<nrow; ++rr)
	for (cc = 0; cc<nx; ++cc) {
	  val = pixmap[(row+rr)*nx+cc];
	  buf[i++] = (uint8) ((val > 255.) ? 255. : (val < 0.) ? 0. : val);
	}
      if (TIFFWriteEncodedStrip(tiffp,strip,buf,i*sizeof(uint8))<0) return EXIT_FAILURE;
    }
    _TIFFfree(buf);
  }
  else return EXIT_FAILURE;

  TIFFWriteDirectory(tiffp);

  return EXIT_SUCCESS;
}


static int _save_tiff_int8(tiffp,pixmap,nx,ny)
     TIFF *tiffp;
     double *pixmap;
     int nx,ny;
{
  const unsigned int directory = 0;
  unsigned int row,rr,cc,vv;
  uint32 nrow;
  tstrip_t strip;
  tsize_t i;
  uint16 photometric = PHOTOMETRIC_MINISBLACK;
  uint32 rowsperstrip = (uint32)-1;
  uint16 samplesperpixel = 1;
  uint16 compression_type = 1;
  uint16 bitspersample = 8;
  double val;

  TIFFSetDirectory(tiffp,directory);
  TIFFSetField(tiffp,TIFFTAG_IMAGEWIDTH,nx);
  TIFFSetField(tiffp,TIFFTAG_IMAGELENGTH,ny);
  TIFFSetField(tiffp,TIFFTAG_ORIENTATION,ORIENTATION_TOPLEFT);
  TIFFSetField(tiffp,TIFFTAG_SAMPLESPERPIXEL,samplesperpixel);
  TIFFSetField(tiffp,TIFFTAG_SAMPLEFORMAT,2);
  TIFFSetField(tiffp,TIFFTAG_BITSPERSAMPLE,bitspersample);
  TIFFSetField(tiffp,TIFFTAG_PLANARCONFIG,PLANARCONFIG_CONTIG);
  TIFFSetField(tiffp,TIFFTAG_PHOTOMETRIC,photometric);
  TIFFSetField(tiffp,TIFFTAG_COMPRESSION,compression_type);
  rowsperstrip = TIFFDefaultStripSize(tiffp,rowsperstrip);
  TIFFSetField(tiffp,TIFFTAG_ROWSPERSTRIP,rowsperstrip);
  TIFFSetField(tiffp,TIFFTAG_FILLORDER,FILLORDER_MSB2LSB);

  int8 *buf = (int8*) _TIFFmalloc(TIFFStripSize(tiffp));
  if (buf) {
    for (row = 0; row<ny; row+=rowsperstrip) {
      nrow = (row + rowsperstrip>ny?ny-row:rowsperstrip);
      strip = TIFFComputeStrip(tiffp,row,0);
      i = 0;
      for (rr = 0; rr<nrow; ++rr)
	for (cc = 0; cc<nx; ++cc) {
	  val = pixmap[(row+rr)*nx+cc];
	  buf[i++] = (int8) ((val > 127.) ? 127. : (val < -128.) ? -128. : val);
	}
      if (TIFFWriteEncodedStrip(tiffp,strip,buf,i*sizeof(int8))<0) return EXIT_FAILURE;
    }
    _TIFFfree(buf);
  }
  else return EXIT_FAILURE;

  TIFFWriteDirectory(tiffp);

  return EXIT_SUCCESS;
}


static _save_tiff_float(tiffp,pixmap,nx,ny)
	TIFF *tiffp;
	double *pixmap;
	int nx,ny;
{
  const unsigned int directory = 0;
  unsigned int row,rr,cc,vv;
  uint32 nrow;
  tstrip_t strip;
  tsize_t i;
  uint16 photometric = PHOTOMETRIC_MINISBLACK;
  uint32 rowsperstrip = (uint32)-1;
  uint16 samplesperpixel = 1;
  uint16 compression_type = 1;
  uint16 bitspersample = 32;

  TIFFSetDirectory(tiffp,directory);
  TIFFSetField(tiffp,TIFFTAG_IMAGEWIDTH,nx);
  TIFFSetField(tiffp,TIFFTAG_IMAGELENGTH,ny);
  TIFFSetField(tiffp,TIFFTAG_ORIENTATION,ORIENTATION_TOPLEFT);
  TIFFSetField(tiffp,TIFFTAG_SAMPLESPERPIXEL,samplesperpixel);
  TIFFSetField(tiffp,TIFFTAG_SAMPLEFORMAT,3);
  TIFFSetField(tiffp,TIFFTAG_BITSPERSAMPLE,bitspersample);
  TIFFSetField(tiffp,TIFFTAG_PLANARCONFIG,PLANARCONFIG_CONTIG);
  TIFFSetField(tiffp,TIFFTAG_PHOTOMETRIC,photometric);
  TIFFSetField(tiffp,TIFFTAG_COMPRESSION,compression_type);
  rowsperstrip = TIFFDefaultStripSize(tiffp,rowsperstrip);
  TIFFSetField(tiffp,TIFFTAG_ROWSPERSTRIP,rowsperstrip);
  TIFFSetField(tiffp,TIFFTAG_FILLORDER,FILLORDER_MSB2LSB);

  float *buf = (float*) _TIFFmalloc(TIFFStripSize(tiffp));
  if (buf) {
    for (row = 0; row<ny; row+=rowsperstrip){
      nrow = (row + rowsperstrip>ny?ny-row:rowsperstrip);
      strip = TIFFComputeStrip(tiffp,row,0);
      i = 0;
      for (rr = 0; rr<nrow; ++rr)
	for (cc = 0; cc<nx; ++cc)
	  buf[i++] = (float) pixmap[(row+rr)*nx + cc];
      if (TIFFWriteEncodedStrip(tiffp,strip,buf,i*sizeof(float))<0) return EXIT_FAILURE;
    }
    _TIFFfree(buf);
  }
  else return EXIT_FAILURE;

  TIFFWriteDirectory(tiffp);

  return EXIT_SUCCESS;
}
